@@ -1,16 +1,13 @@
-from __future__ import print_function
-
 import boto3
 import json
 import os
 import random
 import requests
 import time
-import urllib.parse as urlparse
 
 from collections import namedtuple
 from elasticsearch.exceptions import AuthorizationException
-from urllib.parse import urlencode
+from urllib.parse import parse_qs, urlencode, urlparse, urlunparse
 from . import (
     s3_utils,
     es_utils,
@@ -27,6 +24,9 @@ HIGLASS_BUCKETS = ['elasticbeanstalk-fourfront-webprod-wfoutput',
 # TODO (C4-102): Does this need to include CGAP envs? As part of the same list, or as a separate list?
 PRODUCTION_ENVS = ['fourfront-blue', 'fourfront-green']
 
+# S3 URL identifier
+S3_BUCKET_DOMAIN_SUFFIX = '.s3.amazonaws.com'
+
 
 ##################################
 # Widely used metadata functions #
@@ -235,6 +235,12 @@ def authorized_request(url, auth=None, ff_env=None, verb='GET',
     return result
 
 
+def _sls(val):
+    """general helper to check for and strip leading slashes on ids in API fxns
+    """
+    return val.lstrip('/')
+
+
 def get_metadata(obj_id, key=None, ff_env=None, check_queue=False, add_on=''):
     """
     Function to get metadata for a given obj_id (uuid or @id, most likely).
@@ -251,7 +257,7 @@ def get_metadata(obj_id, key=None, ff_env=None, check_queue=False, add_on=''):
     auth = get_authentication_with_server(key, ff_env)
     if check_queue and stuff_in_queues(ff_env, check_secondary=False):
         add_on += '&datastore=database'
-    get_url = '/'.join([auth['server'], obj_id]) + process_add_on(add_on)
+    get_url = '/'.join([auth['server'], _sls(obj_id)]) + process_add_on(add_on)
     # check the queues if check_queue is True
     response = authorized_request(get_url, auth=auth, verb='GET')
     return get_response_json(response)
@@ -269,7 +275,7 @@ def patch_metadata(patch_item, obj_id='', key=None, ff_env=None, add_on=''):
     if not obj_id:
         raise Exception("ERROR getting id from given object %s for the request to"
                         " patch item. Supply a uuid or accession." % obj_id)
-    patch_url = '/'.join([auth['server'], obj_id]) + process_add_on(add_on)
+    patch_url = '/'.join([auth['server'], _sls(obj_id)]) + process_add_on(add_on)
     # format item to json
     patch_item = json.dumps(patch_item)
     response = authorized_request(patch_url, auth=auth, verb='PATCH', data=patch_item)
@@ -368,10 +374,37 @@ def search_result_generator(page_generator):
     """
     Simple wrapper function to return a generator to iterate through item
     results on the search page
+
+    NOTE: Depending on the nature of the page generator, which may involve separate external calls
+    to a resource like elastic search that is not transactionally managed, the data being queried
+    may change between those calls, usually to add (though theoretically even to remove) an element.
+
+    Consider a case where the data to be queried is indexed in elastic search as A,C,E,G,I,K,M.
+    but where a page size of 3 is used with start position 0. That call will return A,C,E. The
+    user may expect G,I on the second page, but before it can be done, suppose an element D is
+    indexed and that the stored data is A,C,D,E,G,I,K,M. Requesting data from start position 0 would
+    now return A,C,D but we already had the first page, so we request data starting at position 3
+    for the second page and get E,G,I.  That means our sequence of return values would be A,C,E,E,G,I,K,M,
+    or, in other words, showing a duplication. To avoid this, we keep track of the IDs we've seen
+    and show only the first case of each element, so A,C,E,G,I,K,M. (We won't see the D but we weren't
+    going to see it anyway, and it wasn't available the time we started, so the timing was already close.)
+
+    Unfortunately, we aren't so lucky for deletion, though that happens more rarely. That will cause
+    an element to fall out. So if we have A,C,E,G,I,K,M and C is deleted between the first and second call,
+    getting us A,C,E first, and then on the second call when the data is A,E,G,I,K,M we get I,K,M, we'll
+    get the sequence A,C,E,I,K,M and will have missed legitimate element G. There is little to do with this
+    without restarting (which might not terminate or might be O(n^2) in worst case). But deletion is unusual.
     """
+    items_seen = set()
     for page in page_generator:
-        for res in page:
-            yield res
+        for item in page:
+            if isinstance(item, dict):
+                item_uuid = item.get('uuid')
+                if item_uuid:
+                    if item_uuid in items_seen:
+                        continue
+                    items_seen.add(item_uuid)
+            yield item
 
 
 def search_metadata(search, key=None, ff_env=None, page_limit=50, is_generator=False):
@@ -383,7 +416,6 @@ def search_metadata(search, key=None, ff_env=None, page_limit=50, is_generator=F
     Either takes a dictionary form authentication (MUST include 'server')
     or a string fourfront-environment.
     """
-    from urllib.parse import urlparse
     auth = get_authentication_with_server(key, ff_env)
     if search.startswith('/'):
         search = search[1:]
@@ -392,15 +424,23 @@ def search_metadata(search, key=None, ff_env=None, page_limit=50, is_generator=F
         search_url = '/'.join([auth['server'], search])
     else:
         search_url = search  # assume full url is correct
-    search_generator = get_search_generator(search_url, auth=auth, page_limit=page_limit)
+    page_generator = get_search_generator(search_url, auth=auth, page_limit=page_limit)
     if is_generator:
         # yields individual items from search result
-        return search_result_generator(search_generator)
+        return search_result_generator(page_generator)
     else:
         # return a list of all search results
         search_res = []
-        for page in search_generator:
-            search_res.extend(page)
+        items_seen = set()
+        for page in page_generator:
+            for item in page:
+                if isinstance(item, dict):
+                    item_uuid = item.get('uuid')
+                    if item_uuid:
+                        if item_uuid in items_seen:
+                            continue
+                        items_seen.add(item_uuid)
+                search_res.append(item)
         return search_res
 
 
@@ -408,7 +448,7 @@ def get_item_facets(item_type, key=None, ff_env=None):
     """
     Gets facet query string information ie: mapping from facet to query string
     """
-    resp = get_metadata('/profiles/' + item_type + '.json', key=key, ff_env=ff_env)
+    resp = get_metadata('profiles/' + item_type + '.json', key=key, ff_env=ff_env)
     facets = {}
     for query_url, info in resp.get('facets', {}).items():
         facets[info['title']] = query_url
@@ -426,7 +466,7 @@ def get_item_facet_values(item_type, key=None, ff_env=None):
     for that facet mapping to the count for that value
     format: {'Project': {'4DN': 2, 'Other': 6}, 'Lab': {...}}
     """
-    resp = get_metadata('/search/?type=' + item_type, key=key, ff_env=ff_env)['facets']
+    resp = get_metadata('search/?type=' + item_type, key=key, ff_env=ff_env)['facets']
     facets = {}
     for facet in resp:
         name = facet['title']
@@ -459,7 +499,7 @@ def faceted_search(key=None, ff_env=None, item_type=None, **kwargs):
     """
     item_facets = kwargs.get('item_facets', None)
     item_type = 'ExperimentSetReplicate' if item_type is None else item_type
-    search = '/search/?type=' + item_type
+    search = 'search/?type=' + item_type
     if item_facets is None:
         item_facets = get_item_facets(item_type, key=key, ff_env=ff_env)
     for facet, values in kwargs.items():
@@ -535,7 +575,7 @@ def fetch_files_qc_metrics(data, associated_files=None,
                                           'source_file': source_file,
                                           'source_file_type': source_file_type
                                           }
-                                        }
+                            }
                             qc_metrics.update(qc_info)
 
                     else:
@@ -551,7 +591,7 @@ def fetch_files_qc_metrics(data, associated_files=None,
                                       'source_file': source_file,
                                       'source_file_type': source_file_type
                                       }
-                                    }
+                        }
                         qc_metrics.update(qc_info)
     return qc_metrics
 
@@ -868,7 +908,7 @@ def get_schema_names(key=None, ff_env=None):
     """
     auth = get_authentication_with_server(key, ff_env)
     schema_name = {}
-    profiles = get_metadata('/profiles/', key=auth, add_on='frame=raw')
+    profiles = get_metadata('profiles/', key=auth, add_on='frame=raw')
     for key, value in profiles.items():
         # skip abstract types
         if value.get('isAbstract') is True:
@@ -1224,6 +1264,26 @@ def stuff_in_queues(ff_env, check_secondary=False):
     return stuff_in_queue
 
 
+def fetch_network_ids(subnet_names, security_group_names):
+    """ Takes lists of subnet and security group names, and fetches their corresponding ids from AWS."""
+    ec2_client = boto3.client('ec2')
+    subnet_ids = []
+    security_group_ids = []
+    for i in subnet_names:
+        response = ec2_client.describe_subnets()
+        subnet_id = i
+        subnet_ids.append(subnet_id)
+    for i in security_group_names:
+        response = ec2_client.describe_security_groups(  # could possibly be done with just one query
+            Filters=[
+                dict(Name='group-name', Values=[i])
+            ]
+        )
+        security_group_id = response['SecurityGroups'][0]['GroupId']
+        security_group_ids.append(security_group_id)
+    return subnet_ids, security_group_ids
+
+
 def get_response_json(res):
     """
     Very simple function to return json from a response or raise an error if
@@ -1251,12 +1311,12 @@ def process_add_on(add_on):
 
 def get_url_params(url):
     """
-    Returns a dictionary of url params using urlparse.parse_qs.
+    Returns a dictionary of url params using parse_qs.
     Example: get_url_params('<server>/search/?type=Biosample&limit=5') returns
     {'type': ['Biosample'], 'limit': '5'}
     """
-    parsed_url = urlparse.urlparse(url)
-    return urlparse.parse_qs(parsed_url.query)
+    parsed_url = urlparse(url)
+    return parse_qs(parsed_url.query)
 
 
 def update_url_params_and_unparse(url, url_params):
@@ -1266,8 +1326,8 @@ def update_url_params_and_unparse(url, url_params):
     """
     # Note: Although it upsets linting tools, ._replace() is an advertised interface of url.parse -kmp 17-Oct-2020
     #       See https://docs.python.org/3/library/urllib.parse.html
-    parsed_url = urlparse.urlparse(url)._replace(query=urlencode(url_params, True))  # noQA
-    return urlparse.urlunparse(parsed_url)
+    parsed_url = urlparse(url)._replace(query=urlencode(url_params, True))  # noQA
+    return urlunparse(parsed_url)
 
 
 def convert_param(parameter_dict, vals_as_string=False):
@@ -1317,3 +1377,20 @@ def dump_results_to_json(store, folder):
         filename = folder + '/' + a_type + '.json'
         with open(filename, 'w') as outfile:
             json.dump(store[a_type], outfile, indent=4)
+
+
+def parse_s3_bucket_and_key_url(url: str) -> (str, str):
+    """ Parses the given s3 URL into its pair of bucket, key
+        Note that this function works the way it does because of how these
+        urls end up in our database. Eventually we should clean this up.
+        Format:
+            https://s3.amazonaws.com/cgap-devtest-main-application-cgap-devtest-wfout/GAPFI1HVXJ5F/fastqc_report.html
+            https://cgap-devtest-main-application-tibanna-logs.s3.amazonaws.com/41c2fJDQcLk3.metrics/metrics.html
+    """
+    parsed_url = urlparse(url)
+    if parsed_url.hostname.endswith(S3_BUCKET_DOMAIN_SUFFIX):
+        bucket = parsed_url.hostname[:-len(S3_BUCKET_DOMAIN_SUFFIX)]
+        key = parsed_url.path.lstrip('/')
+    else:
+        [bucket, key] = parsed_url.path.lstrip('/').split('/', 1)
+    return bucket, key
